#include <stdlib.h>
#include <limits.h>
#include <errno.h>

#include "ccbcommon.h"
#include "ccbconstants.h"
#include "ccbunsigned.h"

/*
 * Parameterize the number of long integers needed to store the number
 * of nanoseconds stored in a CCBInterval.
 */
#define LONG_PER_INTERVAL 2

/*
 * When an interval is converted from (seconds,nanoseconds) to
 * nanoseconds the bits of the latter large unsigned integer is stored
 * as an array of two unsigned integers.
 */
typedef unsigned long CCBNanoseconds[LONG_PER_INTERVAL];

/*
 * Convert from a CCBInterval value to a CCBNanoseconds value.
 */
static int ccb_interval_to_ns(const CCBInterval *dt, CCBNanoseconds ns);

/*
 * Convert from a CCBNanoseconds value to a CCBInterval value.
 */
static int ccb_ns_to_interval(const CCBNanoseconds ns, CCBInterval *dt);

/*
 * When the integer quotient of two intervals is computed, it is
 * returned as an extended integer of the following array-type.
 */
typedef struct {
  unsigned long quot[LONG_PER_INTERVAL];  /* The quotient of two intervals */
  CCBInterval rem;                        /* The remainder of dividing two */
                                          /*  intervals. */
} CCBDivInterval;

static int ccb_divide_intervals(CCBInterval *dt1, CCBInterval *dt2,
				CCBDivInterval *d);

/*.......................................................................
 * Scale a time interval by an unsigned long integer.
 *
 * Input:
 *  dt   const CCBInterval *  The time interval to be scaled.
 *  factor   unsigned long    The scale factor to apply.
 * Input/Output:
 *  ans        CCBInterval *  The scaled time interval (note that
 *                            'ans' and 'dt' can point to the same
 *                            variable).
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error (see errno).
 */
int ccb_scale_interval(CCBInterval *dt, unsigned long factor, CCBInterval *ans)
{
  CCBNanoseconds dt_ns;      /* The total number of nanoseconds in 'dt' */
  CCBNanoseconds scaled_ns;  /* dt_ns * factor */
  CCBNanoseconds work;       /* A work array */
/*
 * Check the arguments.
 */
  if(!dt || !ans) {
    errno = EINVAL;
    return 1;
  };
/*
 * Convert the interval into an extended integer that represents the
 * total number of nanoseconds in the interval.
 */
  if(ccb_interval_to_ns(dt, dt_ns))
    return 1;
/*
 * Scale the interval by the requested amount.
 */
  if(ccb_ull_multiply(dt_ns, 2, &factor, 1, scaled_ns, 2, work, 2))
    return 1;
/*
 * Convert back from nanoseconds to seconds and nanoseconds.
 */
  if(ccb_ns_to_interval(scaled_ns, ans))
    return 1;
  return 0;
}

/*.......................................................................
 * Add two time intervals.
 *
 * Input:
 *  dt1  const CCBInterval *  The first of the time intervals to added.
 *  dt2  const CCBInterval *  The second of the time intervals to be added.
 * Input/Output:
 *  sum        CCBInterval *  The sum of the intervals. Note that (note
 *                            that 'sum' and 'dt1' can point to the same
 *                            variable).
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error (see errno).
 */
int ccb_add_intervals(CCBInterval *dt1, const CCBInterval *dt2,
		      CCBInterval *sum)
{
  CCBNanoseconds ns1;    /* The total number of nanoseconds in dt1 */
  CCBNanoseconds ns2;    /* The total number of nanoseconds in dt2 */
  CCBNanoseconds ns_sum; /* The sum of ns1 and ns2 */
/*
 * Check the arguments.
 */
  if(!dt1 || !dt2 || !sum) {
    errno = EINVAL;
    return 1;
  };
/*
 * Convert the intervals to extended integers that represent the
 * total number of nanoseconds in each interval.
 */
  if(ccb_interval_to_ns(dt1, ns1) ||
     ccb_interval_to_ns(dt2, ns2))
    return 1;
/*
 * Add the two intervals.
 */
  if(ccb_ull_add(ns1, LONG_PER_INTERVAL, ns2, LONG_PER_INTERVAL,
		 ns_sum, LONG_PER_INTERVAL))
    return 1;
/*
 * Conver the sum back to an interval.
 */
  if(ccb_ns_to_interval(ns_sum, sum))
    return 1;
  return 0;
}

/*.......................................................................
 * Subtract a small time interval from a larger time interval.
 *
 * Input:
 *  dt1  const CCBInterval *  The larger time-interval.
 *  dt2  const CCBInterval *  The time interval to subtract from dt1.
 * Input/Output:
 *  dif        CCBInterval *  The difference dt1-dt2. Note that (note
 *                            that 'dif' and 'dt1' can point to the same
 *                            variable).
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error (see errno).
 */
int ccb_subtract_interval(CCBInterval *dt1, CCBInterval *dt2,
			  CCBInterval *dif)
{
  CCBNanoseconds ns1;     /* The total number of nanoseconds in dt1 */
  CCBNanoseconds ns2;     /* The total number of nanoseconds in dt2 */
  CCBNanoseconds ns_dif;  /* The difference ns1-ns2 */
/*
 * Check the arguments.
 */
  if(!dt1 || !dt2 || !dif) {
    errno = EINVAL;
    return 1;
  };
/*
 * Convert the intervals to extended integers that represent the
 * total number of nanoseconds in each interval.
 */
  if(ccb_interval_to_ns(dt1, ns1) ||
     ccb_interval_to_ns(dt2, ns2))
    return 1;
/*
 * Subtract ns2 from ns1.
 */
  if(ccb_ull_subtract(ns1, LONG_PER_INTERVAL, ns2, LONG_PER_INTERVAL,
		 ns_dif, LONG_PER_INTERVAL))
    return 1;
/*
 * Conver the difference back to an interval.
 */
  if(ccb_ns_to_interval(ns_dif, dif))
    return 1;
  return 0;
}

/*.......................................................................
 * Compare two time intervals.
 *
 * Input:
 *  dt1  const CCBInterval *  The leftmost term of the comparison.
 *  dt2  const CCBInterval *  The rightmost term of the comparison.
 * Output:
 *  return             int    -1:  dt1 < dt2
 *                             0:  dt1 == dt2
 *                             1:  dt1 > dt2
 */
int ccb_compare_intervals(const CCBInterval *dt1, const CCBInterval *dt2)
{
/*
 * Compare the seconds part of the interval, this being the most significant
 * part.
 */
  if(dt1->sec > dt2->sec)
    return 1;
  else if(dt1->sec < dt2->sec)
    return -1;
/*
 * The two seconds terms are equal, so compare the nanoseconds parts.
 */
  else if(dt1->ns > dt2->ns)
    return 1;
  else if(dt1->ns < dt2->ns)
    return -1;
/*
 * Both the seconds and the nano-seconds terms are equal.
 */
  return 0;
}

/*.......................................................................
 * Initialize an interval to 0.
 *
 * Input/Output:
 *  dt     CCBInterval *  The interval to be zeroed.
 */
void ccb_zero_interval(CCBInterval *dt)
{
  if(dt) {
    dt->sec = 0;
    dt->ns = 0;
  };
}

/*.......................................................................
 * Return true if a specified interval is zero.
 *
 * Input:
 *  dt    CCBInterval *  The interval to be checked.
 * Output:
 *  return        int    0 - dt!=0.
 *                       1 - dt==0 || dt==NULL. If dt==NULL, errno will
 *                           be set to EINVAL.
 */
int ccb_interval_is_zero(CCBInterval *dt)
{
/*
 * Check the argument.
 */
  if(!dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * See if both the seconds and nanoseconds fields are zero.
 */
  return dt->sec==0 && dt->ns==0;
}

/*.......................................................................
 * Convert a duration specified as a number of hardware clock ticks to
 * a CCBInterval.
 *
 * Input:
 *  ticks    unsigned long    The duration as a number of clock ticks.
 * Input/Output:
 *  dt         CCBInterval *  The corresponding time interval will be
 *                            assigned to *dt.
 */
void ccb_clock_interval(unsigned long ticks, CCBInterval *dt)
{
/*
 * Split the clock time into seconds and nanoseconds.
 */
  if(dt) {
    dt->sec = ticks / CCB_CLOCK_FREQ;
    dt->ns = (ticks - dt->sec * CCB_CLOCK_FREQ) * CCB_NS_PER_TICK;
  };
}

/*.......................................................................
 * Convert an interval from its split seconds,nanosecond representation
 * to a total number of nanoseconds.
 *
 * Input:
 *  dt  const CCBInterval *  The interval to be converted.
 * Input/Output:
 *  ns     CCBNanoseconds    The number of nanoseconds corresponding
 *                           to the above interval, will be stored in
 *                           ns[].
 * Output:
 *  return            int    0 - OK.
 *                           1 - Error (see errno).
 */
static int ccb_interval_to_ns(const CCBInterval *dt, CCBNanoseconds ns)
{
  CCBNanoseconds work;  /* A work arrays */
/*
 * Get the number of nanoseconds per second.
 */
  const unsigned long ns_per_sec = CCB_NS_PER_SEC;
/*
 * Check the arguments.
 */
  if(!dt || !ns) {
    errno = EINVAL;
    return 1;
  };
/*
 * Scale the seconds component of 'ns', by 1e9, to convert it to
 * nanoseconds, and assign the result to 'ns'.
 */
  if(ccb_ull_multiply(&dt->sec, 1, &ns_per_sec, 1, ns, LONG_PER_INTERVAL,
		      work, LONG_PER_INTERVAL))
    return 1;
/*
 * Add the nanosecond component of 'dt' to 'ns'.
 */
  if(ccb_ull_add(ns, LONG_PER_INTERVAL, &dt->ns, 1, ns, LONG_PER_INTERVAL))
    return 1;
  return 0;
}

/*.......................................................................
 * Convert a total number of nanoseconds to a CCBInterval.
 *
 * Input:
 *  ns  const CCBNanoseconds    The nanoseconds count to be converted.
 *                              Note that zero nano-seconds can be
 *                              expressed by passing 'ns' as NULL.
 * Input/Output:
 *  dt           CCBInterval *  The CCBInterval version of 'ns'.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
static int ccb_ns_to_interval(const CCBNanoseconds ns, CCBInterval *dt)
{
  CCBNanoseconds sec;  /* The seconds component of the interval */
  CCBNanoseconds nsec; /* The nano-seconds component of the interval */
/*
 * Get the number of nanoseconds per second.
 */
  const unsigned long ns_per_sec = CCB_NS_PER_SEC;
/*
 * Check the arguments.
 */
  if(!dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Divide the interval by the number of nanoseconds in an
 * integer. Since this will be an integer division, the resulting
 * quotient will be the number of whole seconds in the interval, and
 * the remainder will be the remaining number of nanoseconds.
 */
  if(ccb_ull_divide(ns, LONG_PER_INTERVAL, &ns_per_sec, 1,
		    sec, LONG_PER_INTERVAL, nsec, LONG_PER_INTERVAL))
    return 1;
/*
 * Assign the return values.
 */
  if(ccb_ull_copy(sec, LONG_PER_INTERVAL, &dt->sec, 1) ||
     ccb_ull_copy(nsec, LONG_PER_INTERVAL, &dt->ns, 1))
    return 1;
  return 0;
}

/*.......................................................................
 * Compute the integer-quotient and interval-remainder of two intervals.
 *
 * Input:
 *  dt1     CCBInterval *  The numerator of the division.
 *  dt2     CCBInterval *  The denominator of the division.
 * Input/Output:
 *  d    CCBDivInterval *  The quotient and remainder of the intervals
 *                         will be recorded in *d.quot and *d.rem
 *                         respectively.
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error (see errno).
 */
static int ccb_divide_intervals(CCBInterval *dt1, CCBInterval *dt2,
				CCBDivInterval *d)
{
  CCBNanoseconds ns1,ns2;  /* The nanosecond equivalents of dt1 and dt2 */
  CCBNanoseconds rem;      /* The remainder of the division */
/*
 * Check the arguments.
 */
  if(!dt1 || !dt2 || !d) {
    errno = EINVAL;
    return 1;
  };
/*
 * Convert the two intervals to nanoseconds.
 */
  if(ccb_interval_to_ns(dt1, ns1) || ccb_interval_to_ns(dt2, ns2))
    return 1;
/*
 * Divide the two numbers.
 */
  if(ccb_ull_divide(ns1, LONG_PER_INTERVAL, ns2, LONG_PER_INTERVAL,
		    d->quot, sizeof(d->quot)/sizeof(d->quot[0]),
		    rem, LONG_PER_INTERVAL))
    return 1;
/*
 * Convert the nanosecond remainder into an interval.
 */
  if(ccb_ns_to_interval(rem, &d->rem))
    return 1;
  return 0;
}

/*.......................................................................
 * Return the settling time of a given combination of simultaneously
 * commanded cal-diode and phase switch transitions.
 *
 * Input:
 *  cnf          const CCBConfig *  The configuration of the CCB.
 *  prev_cal            unsigned    The pre-transition states of the two
 *                                  cal diodes, expressed as a bitwise
 *                                  union of CCBCalDiodes enumerators.
 *  next_cal            unsigned    The post-transition states of the two
 *                                  cal diodes, expressed as a bitwise
 *                                  union of CCBCalDiodes enumerators.
 *  prev_phs            unsigned    The pre-transition states of the two
 *                                  phase-switches, expressed as a bitwise
 *                                  union of CCBPhaseSwitches enumerators.
 *  next_phs            unsigned    The post-transition states of the two
 *                                  phase-switches, expressed as a bitwise
 *                                  union of CCBPhaseSwitches enumerators.
 * Input/Output:
 *  dt               CCBInterval *  The duration of the settling time.
 * Output:
 *  return                   int    0 - OK.
 *                                  1 - Error (see errno).
 */
int ccb_settling_time(const CCBConfig *cnf,
		      unsigned prev_cal, unsigned next_cal,
		      unsigned prev_phs, unsigned next_phs,
		      CCBInterval *dt)
{
  CCBCalDiodeCnf cal;   /* The CCB calibration diode configuration */
  CCBTimingCnf timing;  /* The CCB timing configuration */
  unsigned long delay;  /* The dominant settling time (clock-ticks) */
/*
 * Look up the configuration parameters that we need.
 */
  if(ccb_get_cal_diode_cnf(cnf, &cal) ||
     ccb_get_timing_cnf(cnf, &timing))
    return 1;
/*
 * Clear the unused bits of the cal-diode and phase-switch masks.
 */
  prev_cal &= CCB_ALL_CAL_DIODES;
  next_cal &= CCB_ALL_CAL_DIODES;
  prev_phs &= CCB_ALL_PHASE_SWITCHES;
  next_phs &= CCB_ALL_PHASE_SWITCHES;
/*
 * If there is no change in the states of the cal-diodes, then their
 * contribution to the settling time is zero.
 */
  if(next_cal == prev_cal) {
    delay = 0;
/*
 * Do either of the diodes turn on, and neither turn off?
 */
  } else if(prev_cal == CCB_NO_CAL_DIODES || next_cal == CCB_ALL_CAL_DIODES) {
    delay = timing.diode_rise_dt;
/*
 * Do either of the diodes turn off, and neither turn on?
 */
  } else if(prev_cal == CCB_ALL_CAL_DIODES || next_cal == CCB_NO_CAL_DIODES) {
    delay = timing.diode_fall_dt;
/*
 * To get here, we one diode must be turning on, while the other turns off.
 * Does the rise time of the diode that is turning on, exceed that of the
 * diode that is turning off?
 */
  } else if(timing.diode_rise_dt > timing.diode_fall_dt) {
    delay = timing.diode_rise_dt;
/*
 * To get here, the diode turn-off time must dominate the cal-diode settling
 * time.
 */
  } else {
    delay = timing.diode_fall_dt;
  };
/*
 * Since the rise and fall times of the phase-switches are assumed to be
 * indentical, if either of the phase-switches is changing state, see if the
 * phase-switch settling time dominates the overall settling time.
 */
  if(next_phs != prev_phs && timing.phase_switch_dt > delay)
    delay = timing.phase_switch_dt;
/*
 * Convert the settling time from clock-ticks to the requested time interval.
 */
  ccb_clock_interval(delay, dt);
  return 0;
}

/*.......................................................................
 * Compute the duration of an integration.
 *
 * Input:
 *  cnf       const CCBConfig *  The configuration of the CCB.
 * Input/Output:
 *  dt            CCBInterval *  The integration duration will be
 *                               recorded in *dt.
 * Output:
 *  return                int    0 - OK.
 *                               1 - Error (see errno).
 */
int ccb_integration_duration(const CCBConfig *cnf, CCBInterval *dt)
{
  unsigned long nsample;     /* The number of A/D samples per integration */
  CCBPhaseSwitchCnf phase;   /* The CCB phase-switch configuration */
  CCBTimingCnf timing;       /* The CCB timing configuration */
/*
 * Check the arguments.
 */
  if(!cnf || !dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Look up the configuration parameters that we need.
 */
  if(ccb_get_phase_switch_cnf(cnf, &phase) ||
     ccb_get_timing_cnf(cnf, &timing))
    return 1;
/*
 * Compute the number of samples per integration.
 */
  nsample = timing.integ_period *
    ccb_cycle_length(phase.active_switches) * phase.samp_per_state;
/*
 * Convert this to a time interval.
 */
  ccb_clock_interval(nsample, dt);
  return 0;
}

/*.......................................................................
 * Return the number of measurements per phase-switch cycle, according
 * to the set of phase-switches that are active.
 *
 * Input:
 *  active_switches  unsigned short   (See ccb_set_phase_switch_cnf())
 * Output:
 *  return                 unsigned    The number of measurements per
 *                                     phase-switch cycle.
 */
unsigned ccb_cycle_length(unsigned short active_switches)
{
/*
 * Are both switches active?
 */
  if((active_switches & CCB_ALL_PHASE_SWITCHES) == CCB_ALL_PHASE_SWITCHES)
    return 4;
/*
 * Either one of the switches?
 */
  else if(active_switches & CCB_ALL_PHASE_SWITCHES)
    return 2;
/*
 * Neither of the switches is active.
 */
  return 1;
}

/*.......................................................................
 * Return the expected duration of a fully-described scan of a given
 * number of integration periods.
 *
 * Input:
 *  cnf          const CCBConfig * The configuration of the CCB.
 *  integrations   unsigned long   The number of integrations within
 *                                 the scan.
 * Input/Output:
 *  dt               CCBInterval * The scan duration will be recorded
 *                                 in *dt.
 * Output:
 *  return                   int   0 - OK.
 *                                 1 - Error (see errno).
 */
int ccb_scan_duration(const CCBConfig *cnf, unsigned long integrations,
		      CCBInterval *dt)
{
  CCBInterval integ_dt;     /* The duration of an integration */
/*
 * Check the arguments.
 */
  if(!cnf || !dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Compute the duration of an integration.
 */
  if(ccb_integration_duration(cnf, &integ_dt))
    return 1;
/*
 * Scale the duration of an integration by the number of integrations
 * in the scan.
 */
  if(ccb_scale_interval(&integ_dt, integrations, dt))
    return 1;
/*
 * The total scan duration is now recorded in *dt.
 */
  return 0;
}

/*.......................................................................
 * Return the minimum number of integrations corresponding to a given
 * time interval. Note that either of the n and rdt arguments can be
 * NULL, if the values that they return aren't of interest.
 *
 * Input:
 *  cnf         const CCBConfig *  The configuration of the CCB.
 *  duration        CCBInterval *  The desired duration of the scan.
 * Input/Output:
 *  n             unsigned long *  The number of complete integrations
 *                                 that would fit within the specified
 *                                 time interval, will be assigned to
 *                                 *n.
 *  rdt             CCBInterval *  The remaining time after dividing
 *                                 the specified duration into *n
 *                                 complete integrations, is assigned to
 *                                 *rdt.
 * Output:
 *  return                  int    0 - OK.
 *                                 1 - Error (see errno).
 */
int ccb_integ_per_interval(const CCBConfig *cnf, CCBInterval *duration,
			   unsigned long *n, CCBInterval *rdt)
{
  CCBInterval integ_dt;     /* The duration of an integration */
  CCBTimingCnf timing;      /* The CCB timing configuration */
  CCBDivInterval d;         /* The quotient and remainder of two intervals */
  int i;
/*
 * Check the arguments.
 */
  if(!cnf || !duration) {
    errno = EINVAL;
    return 1;
  };
/*
 * Look up the configuration parameters that we need.
 */
  if(ccb_get_timing_cnf(cnf, &timing))
    return 1;
/*
 * Get the duration of an integration.
 */
  if(ccb_integration_duration(cnf, &integ_dt))
    return 1;
/*
 * Compute the number of integration periods that will fit within
 * specified scan duration.
 */
  if(ccb_divide_intervals(duration, &integ_dt, &d))
    return 1;
/*
 * Make sure that the number of integrations doesn't exceed the length
 * of an unsigned long.  The CCB requirements-specification dictates a
 * minimum integration time of 1ms, so the minimum duration
 * represented by 2^32 integrations is about 50 days. This is much
 * longer than any single scan ought to last.
 */
  for(i=1; i<sizeof(d.quot)/sizeof(d.quot[0]); i++) {
    if(d.quot[i]) {
      errno = ERANGE;
      return 1;
    };
  };
/*
 * Assign the results to the return arguments.
 */
  if(n)
    *n = d.quot[0];
  if(rdt)
    *rdt = d.rem;
  return 0;
}

/*.......................................................................
 * Compute the integration-time of an integration. This is the actual
 * amount of time integrating, and thus excludes samples that are
 * blanked during phase-switch transitions.
 *
 * Input:
 *  cnf       const CCBConfig *  The configuration of the CCB.
 * Input/Output:
 *  dt            CCBInterval *  The integration time will be
 *                               recorded in *dt.
 * Output:
 *  return                int    0 - OK.
 *                               1 - Error (see errno).
 */
int ccb_integration_time(const CCBConfig *cnf, CCBInterval *dt)
{
  CCBPhaseSwitchCnf phase;   /* The CCB phase-switch configuration */
  CCBTimingCnf timing;       /* The CCB timing configuration */
/*
 * Check the arguments.
 */
  if(!cnf || !dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Look up the configuration parameters that we need.
 */
  if(ccb_get_phase_switch_cnf(cnf, &phase) ||
     ccb_get_timing_cnf(cnf, &timing))
    return 1;
/*
 * Compute the duration of a single phase-switch state. Note that
 * there is one ADC sample per FPGA clock tick, so samp_per_state can
 * equally be considered as the the number clock ticks per
 * phase-switch state.
 */
  ccb_clock_interval(phase.samp_per_state, dt);
/*
 * If either of the phase switches are active, subtract the
 * phase-switch transition delay from the above duration of one
 * phase-switch state.
 */
  if(phase.active_switches & CCB_ALL_PHASE_SWITCHES) {
    CCBInterval phase_switch_time;
    ccb_clock_interval(timing.phase_switch_dt, &phase_switch_time);
    if(ccb_subtract_interval(dt, &phase_switch_time, dt))
      return 1;
  };
/*
 * Scale the integration-time of one phase-switch state by the
 * number of states per integration period.
 */
  if(ccb_scale_interval(dt,
	 timing.integ_period * ccb_cycle_length(phase.active_switches), dt))
    return 1;
  return 0;
}

/*.......................................................................
 * Return the number of integrations per cal-diode cycle.
 *
 * Input:
 *  cnf       const CCBConfig *  The configuration of the CCB.
 * Input/Output:
 *  ninteg      unsigned long *  The number of integrations per cal-cycle
 *                               will be assigned to *ninteg.
 * Output:
 *  return                   int    0 - OK.
 *                                  1 - Error (see errno).
 */
int ccb_cal_cycle_length(const CCBConfig *cnf, unsigned long *ninteg)
{
  CCBCalDiodeCnf cal;   /* The CCB calibration diode configuration */
  int i;
/*
 * Check the arguments.
 */
  if(!cnf || !ninteg) {
    errno = EINVAL;
    return 1;
  };
/*
 * Look up the configuration parameters that we need.
 */
  if(ccb_get_cal_diode_cnf(cnf, &cal))
    return 1;
/*
 * Total-up the number of integrations in a complete calibration cycle.
 */
  *ninteg = 0;
  for(i=0; i<cal.ncal; i++)
    *ninteg += cal.diode_times[i];
  return 0;
}

