#ifndef ccbtelemetryclient_h
#define ccbtelemetryclient_h

#include <stdarg.h>
#include "ccbclientlink.h"

/*-----------------------------------------------------------------------*
 * This file defines the client side of the telemetry-link               *
 * communications interface between a CCB manager and a CCB server.      *
 *-----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*
 *              The manager interface to the telemetry link              *
 *-----------------------------------------------------------------------*
 *
 * CCBTelemetryClient objects manage the client side of the telemetry link
 * to a CCB server.
 */
typedef struct CCBTelemetryClient CCBTelemetryClient;

CCBTelemetryClient *new_CCBTelemetryClient(CCBClientLink *client);
CCBTelemetryClient *del_CCBTelemetryClient(CCBTelemetryClient *cc);


/*
 * Register the callback function to be called whenever a given CCB client
 * object receives a monitor data message.
 */
int ccb_tc_monitor_msg_callback(CCBTelemetryClient *tc,
				CCBMonitorMsgFn *fn, void *data);

/*
 * Register the callback function to be called whenever a given CCB client
 * object receives an integrated-data message.
 */
int ccb_tc_integ_msg_callback(CCBTelemetryClient *tc, CCBIntegMsgFn *fn,
			      void *data);


/*
 * Register the callback function to be called whenever a given CCB client
 * object receives a log message.
 */
int ccb_tc_log_msg_callback(CCBTelemetryClient *tc, CCBLogMsgFn *fn,
			    void *data);


/*
 * Respond to the reported writability of the telemetry socket.  Since
 * telemetry clients never send data to CCB servers over the telemetry
 * link, this can only mean the completion of a non-blocking
 * connection.
 */
int ccb_tc_send_telemetry(CCBTelemetryClient *tc);

/*
 * Read up to one command message from the CCB client telemetry socket,
 * and when a complete command-message has been read, deliver it to
 * the manager by calling the appropriate callback function.
 */
int ccb_tc_read_telemetry(CCBTelemetryClient *tc);

/*
 * Initiate a connection to a CCB telemetry server.
 */
int ccb_tc_connect(CCBTelemetryClient *tc, const char *host, int nonblocking);

/*
 * Select between blocking and non-blocking telemetry-client socket I/O.
 */
int ccb_tc_non_blocking_io(CCBTelemetryClient *tc, int on);

/*
 * Return the file-descriptor of the telemetry socket.
 */
int ccb_tc_socket(CCBTelemetryClient *tc);

/*
 * Disconnect from a remote CCB telemetry server.
 */
void ccb_tc_disconnect(CCBTelemetryClient *tc);

/*
 * Enumerate the states that the telemetry client can be in. Note that
 * telemetry data is always expected, and the client must thus always
 * watch for the readability of the telemetry socket, so this
 * requirement isn't enumerated. Also, no messages are ever sent by the
 * client to the server.
 */
typedef enum {
  CCB_TELEM_CLIENT_DOWN,   /* Their is currently no telemetry link */
  CCB_TELEM_CLIENT_PEND,   /* A non-blocking connect is in progress */
  CCB_TELEM_CLIENT_ACTIVE  /* The telemetry client is active */
} CCBTelemClientStatus;

/*
 * Query the current status of the telemetry client.
 */
CCBTelemClientStatus ccb_tc_status(CCBTelemetryClient *tc);

/*
 * Format an error message and use the manager's log-reporting
 * callback to deliver it.
 */
int ccb_tc_vlog_client_msg(CCBTelemetryClient *tc, CCBLogLevel level,
			   unsigned long id, const char *fmt, va_list ap);

#endif
