#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <math.h>

#include <sys/time.h>

#include "ccbcommon.h"
#include "ccbunsigned.h"

/*.......................................................................
 * Return the current date and time (utc).
 *
 * Input/Output:
 *  ts    CCBTimeStamp *  The current UTC date and time will be recorded
 *                        in *ts.
 * Output:
 *  return         int    0 - OK.
 *                        1 - Time not available.
 */
int ccb_get_timestamp(CCBTimeStamp *ts)
{
  struct timeval tv;   /* The current clock time, with up to usec resolution */
/*
 * Check the arguments.
 */
  if(!ts) {
    errno = EINVAL;
    return 1;
  };
/*
 * Get the current calendar time with up to micro-second resolution.
 */
  if(gettimeofday(&tv, NULL))
    return 1;
/*
 * Convert the date/time to a timestamp.
 */
  return ccb_time_to_timestamp(tv.tv_sec, tv.tv_usec*CCB_NS_PER_USEC, ts);
}

/*.......................................................................
 * Convert a date-and-time specified via a standard time_t variable,
 * plus a seconds fraction, to a CCB timestamp.
 *
 * Input:
 *  t            time_t    The date-and-time to be converted in seconds since
 *                         the epoch used by the C library.
 *  ns    unsigned long    The fractional part of the second, expressed in
 *                         nano-seconds.
 * Input/Output:
 *  ts     CCBTimeStamp *  The timestamp will be assigned to the variable
 *                         that 'ts' points at.
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error (see errno).
 */
int ccb_time_to_timestamp(time_t t, unsigned long ns, CCBTimeStamp *ts)
{
  struct tm utc;       /* The current UTC date and time to second resolution */
  long jd;             /* The Julian Day Number */
  int y;               /* The Gregorian year, including the century */
  int m;               /* The month of the year (1-12) */
  int d;               /* The day of the month (1-31) */
/*
 * Check the arguments.
 */
  if(!ts) {
    errno = EINVAL;
    return 1;
  };
/*
 * Convert 't' to UTC.
 */
  if(!gmtime_r(&t, &utc)) {
    errno = ENOSYS;
    return 1;
  };
/*
 * Get local copies of the components of the calendar date with the origins
 * that are needed by the Julian-Day-Number equation.
 */
  y = utc.tm_year + 1900;    /* Year with century included */
  m = utc.tm_mon + 1;        /* Month in range 1-12 */
  d = utc.tm_mday;           /* Day within the month, in the range 1-31 */
/*
 * Compute the Julian Day Number using equation 12.92-1 on page 604 of
 * the Explanatory Supplement to the Astronomical Almanac (1992).
 */
  jd = (1461L * (y + 4800L + (m - 14L) / 12L)) / 4L +
       (367L * (m - 2L - 12L * ((m - 14L) / 12L))) / 12L -
       (3L * ((y + 4900L + (m - 14L) / 12L) / 100L)) / 4L +
       d - 32075L;
/*
 * Compute and record the MJD and time of day in the return argument.
 * Note that whereas in floating point MJD=jd-2400000.5, our
 * time-of-day components are already relative to the MJD origin, so
 * we use the integer expression JD-2400001 without losing the time of
 * day.
 */
  ts->mjd = jd - 2400001;
  ts->sec = utc.tm_sec + 60 * (utc.tm_min + utc.tm_hour * 60);
  ts->ns = ns;
  return 0;
}

/*.......................................................................
 * Convert the time of day within a CCBTimeStamp into hours minutes and
 * seconds.
 *
 * Input:
 *  ts  CCBTimeStamp    The timestamp to convert.
 * Input/Output:
 *  h,m,s   unsigned *  The hours minutes and seconds within the day.
 * Output:
 *  return       int    0 - OK.
 *                      1 = Error.
 */
void ccb_hms_of_timestamp(CCBTimeStamp ts, unsigned *hours, unsigned *mins,
			 unsigned *secs)
{
  int h,m,s;  /* The hours, minutes and seconds to be returned */
  div_t d;    /* The quotient and remainder of a division */
/*
 * Split the seconds representation of the time-of-day into hours minutes
 * and seconds.
 */
  d = div((int)ts.sec, 3600);
  h = d.quot;
  d = div(d.rem, 60);
  m = d.quot;
  s = d.rem;
/*
 * Return the components of the time.
 */
  if(hours)
    *hours = h;
  if(mins)
    *mins = m;
  if(secs)
    *secs = s;
  return;
}

/*.......................................................................
 * Compare two timestamps.
 *
 * Input:
 *  ta   CCBTimeStamp  The first of the timestamps to be compared.
 *  tb   CCBTimeStamp  The second of the timestamps to be compared.
 * Output:
 *  return        int  -1  ta < tb.
 *                      0  ta == tb.
 *                      1  ta > tb.
 */
int ccb_compare_timestamps(CCBTimeStamp ta, CCBTimeStamp tb)
{
/*
 * Compare the Modified Julian day numbers.
 */
  if(ta.mjd < tb.mjd)
    return -1;
  else if(ta.mjd > tb.mjd)
    return 1;
/*
 * Both timestamps have the same date, so compare their seconds time-of-day
 * fields.
 */
  if(ta.sec < tb.sec)
    return -1;
  else if(ta.sec > tb.sec)
    return 1;
/*
 * Both timestamps have the same dates and integral number of seconds,
 * so compare the fractional seconds of the two timestamps.
 */
  if(ta.ns < tb.ns)
    return -1;
  else if(ta.ns > tb.ns)
    return 1;
/*
 * The timestamps are the same.
 */
  return 0;
}

/*.......................................................................
 * Return the amount of time remaining until a given time, or zero if the
 * time has already been reached or passed.
 *
 * Input:
 *  ts   CCBTimeStamp    The target time.
 * Input/Output:
 *  dt    CCBInterval *  The amount of time to remaining will be assigned
 *                       to *dt.
 * Output:
 *  return        int    0 - OK.
 *                       1 - Error.
 */
int ccb_time_until(CCBTimeStamp ts, CCBInterval *dt)
{
  CCBTimeStamp now;   /* The current time */
/*
 * Check the arguments.
 */
  if(!dt) {
    errno = EINVAL;
    return 1;
  };
/*
 * Get the current time.
 */
  if(ccb_get_timestamp(&now))
    return 1;
/*
 * Start with an zero-length interval.
 */
  dt->sec = 0;
  dt->ns = 0;
/*
 * Is the target time still in the future.
 */
  if(ccb_compare_timestamps(now, ts) < 0) {
/*
 * If the dates differ, compute the corresponding difference in seconds,
 * using arbitrary precision integer functions to detect overflows.
 */
    if(ts.mjd != now.mjd) {
      unsigned long d = ts.mjd - now.mjd;
      unsigned long factor = CCB_SEC_PER_DAY;
      unsigned long work;
      if(ccb_ull_multiply(&d, 1, &factor, 1, &dt->sec, 1, &work, 1))
	return 1;
    };
/*
 * Add in the difference in seconds.
 */
    if(ts.sec > now.sec) {
      unsigned long ds = ts.sec - now.sec;
      if(ccb_ull_add(&dt->sec, 1, &ds, 1, &dt->sec, 1))
	return 1;
    } else if(ts.sec < now.sec) {
      unsigned long ds = now.sec - ts.sec;
      if(ccb_ull_subtract(&dt->sec, 1, &ds, 1, &dt->sec, 1))
	return 1;
    };
/*
 * Finally, add in the difference in nano-seconds.
 */
    if(ts.ns > now.ns) {
      unsigned long du = ts.ns - now.ns;
      unsigned long ds = du / CCB_NS_PER_SEC; /* Whole seconds */
      if(ccb_ull_add(&dt->sec, 1, &ds, 1, &dt->sec, 1))
	return 1;
      dt->ns = du - ds * CCB_NS_PER_SEC;
    } else if(ts.ns < now.ns) {
      unsigned long du = now.ns - ts.ns;
      unsigned long ds = du / CCB_NS_PER_SEC; /* Whole seconds */
      du -= ds * CCB_NS_PER_SEC;
      if(du == 0) {
	dt->ns = 0;
      } else {
	ds++;
	dt->ns = CCB_NS_PER_SEC - du;
      };
      if(ds && ccb_ull_subtract(&dt->sec, 1, &ds, 1, &dt->sec, 1))
	return 1;
    };
  };
  return 0;
}

/*.......................................................................
 * Add a time interval to a timestamp.
 *
 * Input:
 *  ta  const CCBTimeStamp *  The timestamp to add to.
 *  dt   const CCBInterval *  The time interval to add.
 * Input/Output:
 *  tb        CCBTimeStamp *  A pointer to the timestamp to be added to.
 *                            Note that tb can be the same as ta.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_add_to_timestamp(const CCBTimeStamp *ta, const CCBInterval *dt,
			 CCBTimeStamp *tb)
{
  CCBInterval tod;  /* The time of day split into seconds and nanoseconds */
/*
 * Check the arguments.
 */
  if(!tb) {
    errno = EINVAL;
    return 1;
  };
/*
 * Copy the time-of-day into an interval, so that we can pass it to
 * CCBInterval method functions.
 */
  tod.sec = ta->sec;
  tod.ns = ta->ns;
/*
 * Add the specified interval to the time-of-day.
 */
  if(ccb_add_intervals(&tod, dt, &tod))
    return 1;
/*
 * Copy the existing timestamp into the output argument.
 */
  if(ta != tb)
    *tb = *ta;
/*
 * If the time-of-day exceeds the length of a day, increment the date
 * component of the timestamp and update the time-of-day accordingly.
 */
  if(tod.sec >= CCB_SEC_PER_DAY) {
    tb->mjd += tod.sec / CCB_SEC_PER_DAY;
    tod.sec %= CCB_SEC_PER_DAY;
  };
/*
 * Record the new time-of-day in the return value.
 */
  tb->sec = tod.sec;
  tb->ns = tod.ns;
  return 0;
}
